import torch

class BaseDistiller:
    def __init__(self, teacher, student, criterion):
        self.teacher = teacher
        self.student = student
        self.criterion = criterion

    def __call__(self):
        raise NotImplementedError

    def get_callback(self):
        pass


class OutputDistiller(BaseDistiller):
    '''Distilling the Knowledge in a Neural Network, NIPSW 2014
    https://arxiv.org/abs/1503.02531
    '''

    def __init__(self, teacher, student, criterion):
        super().__init__(teacher, student, criterion)

    def __call__(self, x):
        with torch.no_grad():
            y_t = self.teacher(x)
        y_s = self.student(x)
        loss = self.criterion(y_s, y_t)
        return loss


class FeatureDistiller(BaseDistiller):
    def __init__(self, teacher, student, criterion):
        super().__init__(teacher, student, criterion)

    def __call__(self, x):
        pass


class SimilarityDistiller(BaseDistiller):
    '''Similarity-Preserving Knowledge Distillation, ICCV 2019
    https://arxiv.org/abs/1907.09682
    '''

    def __init__(self, teacher, student, criterion):
        super().__init__(teacher, student, criterion)

    def __call__(self, x):
        pass
