import sys
sys.path.append('../../..')

from knnslim.utils import enums
from knnslim.prune.pruning_method import base


class WeightPruningMethod(base.UnstructuredPruningMethod):
    def __init__(self, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_engine=pruning_engine)

    def prune(self, weight_pt, amount, criterion, importance_scores=None):
        device = weight_pt.device
        weight = self._t(weight_pt)

        # get the number of pruning weights
        nweights = self._get_nweights(weight)
        nweights_toprune = self._get_nweights_toprune(nweights, amount)

        # calculate the importance scores of all weights
        if not importance_scores:
            if criterion:
                self.set_pruning_criterion(criterion)
            importance_scores = self.pruning_criterion(weight=weight)
        self._validate_importance_scores(weight, importance_scores)

        # obtain the index of pruning weights
        pruning_weights = self._bottomk(importance_scores.reshape(-1), nweights_toprune)

        # obtain the pruning mask and prune the weight
        mask = self._ones((nweights,), dtype=bool)
        mask[pruning_weights] = False
        mask = mask.reshape(weight.shape)
        weight_masked = self._apply_mask(weight, mask)

        weight_masked_pt = self._u(weight_masked, device)
        mask_pt = self._u(mask, device)
        return weight_masked_pt, mask_pt
