import torch
import torch.utils.data
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.autograd import Variable
from torch.nn.parameter import Parameter
#import torchvision.ops as ops

try:
    from export_configs import forward_config
except:
    pass

USING_MASK_RCNN = 0
if USING_MASK_RCNN:
    from common_utils.mask_rcnn.modeling.FPN import fpn_rpn_outputs
    from common_utils.mask_rcnn.operation3 import RoiSelection
    from common_utils.mask_rcnn.modeling.fast_rcnn_heads import roi_2mlp_head, fast_rcnn_outputs
    from common_utils.mask_rcnn.modeling.mask_rcnn_heads import mask_rcnn_outputs, mask_rcnn_fcn_head_v1upXconvs
else:
    class fpn_rpn_outputs():
        def __init__(self):
            pass
        def forward(self, x):
            pass

    class RoiSelection():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class roi_2mlp_head():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class fast_rcnn_outputs():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class mask_rcnn_outputs():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class mask_rcnn_fcn_head_v1upXconvs():
        def __init__(self):
            pass

        def forward(self, x):
            pass

USE_WIREFRAME=0
if USE_WIREFRAME:
    from common_utils.wire_frame import ProposalLines
    from common_utils.wire_frame import LineJuncPick
    from common_utils.wire_frame import LineJuncDecoder
    from common_utils.wire_frame import LOIPooling
else:
    class ProposalLines():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class LineJuncPick():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class LineJuncDecoder():
        def __init__(self):
            pass

        def forward(self, x):
            pass

    class LOIPooling():
        def __init__(self):
            pass

        def forward(self, x):
            pass

#==================================================================================================================

class Assign_Conv2D(nn.Conv2d):

    def _conv_forward(self, input, weight):
        return F.conv2d(input, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)

    def forward(self, inputs, assign_weight):
        return self._conv_forward(inputs, self.weight)


#== add new op after 2020.04.25 for PointRend
class TopK(nn.Module):
    def __init__(self, k, dim, largest=True, outIdx=1):
        super(TopK, self).__init__()
        self.k = k
        self.dim = dim
        self.largest = largest
        self.outIdx = outIdx # 0, 1

    def forward(self, inputs):
        out = torch.topk(inputs, self.k, dim=self.dim, largest=self.largest)
        return out[self.outIdx]

class uncertain_point(nn.Module):
    def __init__(self , num_points):
        super(uncertain_point, self).__init__()
        self.num_points = num_points

    def forward(self, uncertainty_map):
        R, _, H, W = uncertainty_map.shape
        h_step = 1.0 / float(H)
        w_step = 1.0 / float(W)

        num_points = min(H * W, self.num_points)
        point_indices = torch.topk(uncertainty_map.view(R, H * W), k=num_points, dim=1)[1]
        point_coords = torch.zeros(R, num_points, 2, dtype=torch.float, device=uncertainty_map.device)
        point_coords[:, :, 0] = w_step / 2.0 + (point_indices % W).to(torch.float) * w_step
        point_coords[:, :, 1] = h_step / 2.0 + (point_indices // W).to(torch.float) * h_step
        return point_indices, point_coords

class PointSample(nn.Module):
    def __init__(self,
                 data_in_type='float',
                 data_out_type='float',
                 min=0.0, max=4.0):

        super(PointSample, self).__init__()
        self.data_in_type = data_in_type # float / fixed
        self.data_out_type = data_out_type # float / fixed
        self.minv = min
        self.maxv = max

    def forward(self, inputs):
        add_dim = False
        input, point_coords = inputs
        if point_coords.dim() == 3:
            add_dim = True
            point_coords = point_coords.unsqueeze(2)
        try:
            output = F.grid_sample(input, 2.0 * point_coords - 1.0, align_corners=True)
        except:
            output = F.grid_sample(input, 2.0 * point_coords - 1.0)
        if add_dim:
            output = output.squeeze(3)
        if self.data_out_type == 'fixed':
            output.data[:] = torch.round(output.data[:] * (255. / self.maxv)) * (self.maxv / 255.)
        return output

class PointSampleQuant(nn.Module):
    def __init__(self, min=0.0, max=4.0):
        super(PointSampleQuant, self).__init__()
        self.minv = min
        self.maxv = max

    def forward(self, inputs):
        add_dim = False
        input, point_coords = inputs
        if point_coords.dim() == 3:
            add_dim = True
            point_coords = point_coords.unsqueeze(2)
        try:
            output = F.grid_sample(input, 2.0 * point_coords - 1.0, align_corners=True)
        except:
            output = F.grid_sample(input, 2.0 * point_coords - 1.0)
        if add_dim:
            output = output.squeeze(3)
        output.data[:] = torch.round(output.data[:] * (255. / self.maxv)) * (self.maxv / 255.)
        return output

class Coords(nn.Module):
    def __init__(self, H, W, num_points):
        super(Coords, self).__init__()
        self.H = H
        self.W = W
        self.h_step = 1.0 / H
        self.w_step = 1.0 / W
        self.num_points = num_points

    def forward(self, inputs):
        # h_step = 1.0 / float(self.H)
        # w_step = 1.0 / float(self.W)
        point_coords, point_indices = inputs
        point_coords[:, :, 0] = self.w_step / 2.0 + (point_indices % self.W).to(torch.float) * self.w_step
        point_coords[:, :, 1] = self.h_step / 2.0 + (point_indices // self.W).to(torch.float) * self.h_step
        return point_coords

class BboxConvert(nn.Module):
    def __init__(self, n, c, pad):
        super(BboxConvert, self).__init__()
        self.num_N = n
        self.num_C = c
        self._pad = pad

    def forward(self, x):
        rects_use = x
        for idx_i in range(self.num_N):
            for idx_j in range(self.num_C):
                x1, y1, x2, y2 = rects_use[idx_i, 4*idx_j:4*(idx_j+1)]
                x1, y1 = max(0, x1-self._pad), max(0, y1-self._pad)
                x2, y2 = min(1, x2+self._pad), min(1, y2+self._pad)
                if x1>=x2 or y1>=y2:
                    x1, y1, x2, y2 = 0, 0, 1, 1
                rects_use[idx_i, 4*idx_j] = x1
                rects_use[idx_i, 4*idx_j+1] = y1
                rects_use[idx_i, 4*idx_j+2] = x2
                rects_use[idx_i, 4*idx_j+3] = y2
        return rects_use

# inputs = (feature_map, rois); rois=(1, 5), rois[0, :]=[0, x1, y1, x2, y2], 归一化数值
class ROIAlign(nn.Module):
    def __init__(self, output_size_h=32, output_size_w=32, spatial_scale=128, sampling_ratio=-1):
        super(ROIAlign, self).__init__()
        raise('not implemented')
        self.output_size_h = output_size_h
        self.output_size_w = output_size_w
        self.spatial_scale = spatial_scale
        self.sampling_ratio = sampling_ratio

    def forward(self, inputs):
        feature_map, rois = inputs[0], inputs[1]
        rois_feature = ops.roi_align(feature_map, rois, [self.output_size_h, self.output_size_w],
                                    spatial_scale=self.spatial_scale, sampling_ratio=self.sampling_ratio)
        return rois_feature

#
class MinOp(nn.Module):
    def __init__(self, dim=None, keepdim=True):
        super(MinOp, self).__init__()

        if not isinstance(dim, int):
            raise ValueError("attr dim must be int, not {}.".format(type(dim)))
        elif dim == 0:
            raise ValueError("not support dim = 0.")

        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x):
        # if self.dim:
        #     out = torch.max(x, dim=self.dim, keepdim=self.keepdim)[0]
        # else:
        #     out = torch.max(x, keepdim=self.keepdim)
        out = torch.min(x, dim=self.dim, keepdim=self.keepdim)[0]
        return out

class MaxOp(nn.Module):
    def __init__(self, dim=None, keepdim=True):
        super(MaxOp, self).__init__()

        if not isinstance(dim, int):
            raise ValueError("attr dim must be int, not {}.".format(type(dim)))
        elif dim == 0:
            raise ValueError("not support dim = 0.")

        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x):
        # if self.dim:
        #     out = torch.max(x, dim=self.dim, keepdim=self.keepdim)[0]
        # else:
        #     out = torch.max(x, keepdim=self.keepdim)
        out = torch.max(x, dim=self.dim, keepdim=self.keepdim)[0]
        return out

class MeanOp(nn.Module):
    def __init__(self, dim=None, keepdim=False):
        super(MeanOp, self).__init__()

        if not isinstance(dim, int):
            raise ValueError("attr dim must be int, not {}.".format(type(dim)))
        elif dim == 0:
            raise ValueError("not support dim = 0.")

        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x):
        # if self.dim:
        #     out = torch.mean(x, dim=self.dim, keepdim=self.keepdim)[0]
        # else:
        #     out = torch.mean(x, keepdim=self.keepdim)
        out = torch.mean(x, dim=self.dim, keepdim=self.keepdim)
        out = out.unsqueeze(self.dim)
        return out

class NormOp(nn.Module):

    '''
    p = L1, L2
    dim = [C, None, None],
    data_in_type= "fixed, "float"
    '''

    def __init__(self, data_in_type="float", p='L2',
                 dim=None, calcu_mean=False, min_val=1e-5):

        super(NormOp, self).__init__()

        assert data_in_type in ['fixed', 'float']

        self.p = p
        self.dim = dim
        self.min_val = min_val
        self.caclu_mean = calcu_mean
        self.data_in_type = data_in_type

        if dim is None or not isinstance(dim, list):
            raise ValueError("type error")

    def forward(self, x):
        # if self.dim == None:
        #     dim = list(np.arange(len(x.size())))
        #     norm = torch.norm(x, p=self.p, dim=dim[1:], keepdim=False)
        #     if self.caclu_mean:
        #         mean = torch.mean(torch.pow(x, 2), dim=dim[1:], keepdim=False)
        #         norm = norm * mean
        #     norm = norm.unsqueeze(1)
        # else:
        #     norm = torch.norm(x, p=self.p, dim=self.dim, keepdim=False)
        #     if self.caclu_mean:
        #         mean = torch.mean(torch.pow(x, 2), dim=self.dim, keepdim=False)
        #         norm = norm * mean
        #     norm = norm.unsqueeze(self.dim)

        if self.dim == None:
            dim = list(np.arange(len(x.size())))
            dim = dim[1:]
        else:
            dim = self.dim

        if self.p == 'L2':
            if self.caclu_mean:
                # norm = torch.sqrt(torch.mean(torch.pow(x, 2), dim=dim, keepdim=False))
                norm = torch.pow(x, 2).mean(dim=dim, keepdim=False).sqrt()
            else:
                norm = torch.norm(x, p=None, dim=dim, keepdim=False)
            norm = norm.unsqueeze(1)

        norm = norm.clamp_min(self.min_val)
        out = x / norm
        return out

class CompareOp(nn.Module):
    def __init__(self, mode, value):
        super(CompareOp, self).__init__()

        self.mode = mode # "==", "<", ">", "<=", ".>="
        self.value = value

    def forward(self, x):
        if self.mode == "==":
            out = (x == self.value)
            return out.float()

class Split(nn.Module):
    def __init__(self, split_size_or_sections, dim, output_size=None):
        super(Split, self).__init__()

        self.split_size_or_sections = split_size_or_sections
        self.dim = dim
        self.output_size = output_size

    def foward(self, x):
        out = torch.split(x, self.split_size_or_sections, self.dim)
        return out

#===============================


class AvgQuant(nn.Module):
    dump_patches = True
    def __init__(self):
        super(AvgQuant, self).__init__()
    def forward(self, x1, x2):
        out = (x1 + x2)/2
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class HardQuant(nn.Module):
    dump_patches = True
    def __init__(self, min_val = 0.0, max_val = 4.0, inplace = False):
        super(HardQuant, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.delta = self.max_val - self.min_val
        self.inplace = inplace
    def forward(self, x):
        out = F.hardtanh(x, self.min_val, self.max_val, self.inplace)
        if self.min_val == 0:  # 只有最小为0时，需要量化，否则输出是浮点
            out.data[:] = torch.round(out.data[:] * (255. / self.delta)) * (self.delta / 255.)
        return out

# deprecated: same with HardQuant above
class QuantizedHardtanh(nn.Module):
    dump_patches = True
    def __init__(self, min_val = 0, max_val = 4, inplace = False):
        super(QuantizedHardtanh, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.delta = self.max_val - self.min_val
        self.inplace = inplace
    def forward(self, x):
        out = F.hardtanh(x, self.min_val, self.max_val, self.inplace)
        out.data[:] = torch.round(out.data[:] * (255. / self.delta)) * (self.delta / 255.)
        return out

# Quant Model should use this. Float model just use * directly
class SliceMul(nn.Module):
    dump_patches = True
    """
    x1: NxCxHxW feature map
    x2: NxCx1x1 SE layer output
    """
    def __init__(self):
        super(SliceMul, self).__init__()

    def forward(self, x1, x2):
        out = x1 * x2
        # Note: divide by 256, consistance with Modile-Neon
        out.data[:] = torch.floor(out.data[:] * (255. / 256.) * (255. / 4.)) * (4. / 255.)
        return out

class SliceAddQuant(nn.Module):
    dump_patches = True
    """
    x1: NxCxHxW feature map
    x2: NxCx1x1
    """
    def __init__(self):
        super(SliceAddQuant, self).__init__()

    def forward(self, x1, x2):
        out = (x1 + x2)
        # Note: divide by 256, consistance with Modile-Neon
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

# Quant Model should use this. Float model just use * directly
class SliceMulParams(nn.Module):
    dump_patches = True
    """
    x1: NxCxHxW feature map
    x2: NxCx1x1 SE layer output
    """
    def __init__(self, out_channels, weights):
        super(SliceMulParams, self).__init__()
        self.out_channels = out_channels
        self.weights = torch.from_numpy(weights.astype(np.float32).reshape(1, out_channels, 1, 1))

    def forward(self, x1):
        out = x1 * self.weights.expand(-1, self.out_channels, 1, 1)
        return out

# deprecated: MulQuant should not used directly in Quant Models
class MulQuant(nn.Module):
    dump_patches = True
    def __init__(self):
        super(MulQuant, self).__init__()

    def forward(self, x1, x2):
        out = x1 * x2
        # divide by 256, consistance with Modile-Neon
        out = torch.clamp(out, 0, 4)
        out.data[:] = torch.floor(out.data[:] * (255. / 256.) * (255. / 4.)) * (4. / 255.)
        return out

# Consider to use AvgQuant in Quant Models
# class AddQuant(nn.Module):
#     dump_patches = True
#     def __init__(self, max_val=4.0):
#         super(AddQuant, self).__init__()
#         self.max_val = max_val
#     def forward(self, x1, x2):
#         out = x1 + x2
#         out = torch.clamp(out, 0, self.max_val)
#         return out

class AvgPoolQuant(nn.AvgPool2d):
    dump_patches = True
    def __init__(self, kernel_size, stride=None, padding=0, ceil_mode=False,
                 count_include_pad=True):
        super(AvgPoolQuant, self).__init__(kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode,
                                      count_include_pad=count_include_pad)

    def forward(self, x):
        out = F.avg_pool2d(x, self.kernel_size, self.stride,
                           self.padding, self.ceil_mode, self.count_include_pad)
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class AdaptiveAvgPoolQuant(nn.AdaptiveAvgPool2d):
    dump_patches = True
    def __init__(self, output_size):
        super(AdaptiveAvgPoolQuant, self).__init__(output_size)
    def forward(self, x):
        out = F.adaptive_avg_pool2d(x, self.output_size)
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class UpsampleQuant(nn.Upsample):
    dump_patches = True
    def __init__(self, size=None, scale_factor=None, max=4, mode='nearest', align_corners=True):
        super(UpsampleQuant, self).__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corners = align_corners
        self.maxv = max

    def forward(self, x):
        if self.mode == "nearest":
            out = F.upsample(x, self.size, self.scale_factor, self.mode)
        else:
            out = F.upsample(x, self.size, self.scale_factor, self.mode, self.align_corners)

        out.data[:] = torch.round(out.data[:] * (255. / self.maxv)) * (self.maxv / 255.)
        return out

class Conv2dQuant(nn.Conv2d):
    dump_patches = True
    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=1, padding=0, dilation=1, groups=1, bias=False):
        super(Conv2dQuant, self).__init__(in_channels, out_channels,
                                          kernel_size, stride, padding,
                                          dilation, groups=groups, bias=bias)
        self.register_buffer('weight_copy', torch.zeros(self.weight.size()))
        self.computation = 0

    def get_computation(self, x):
        feature_size = x.size()[-1]
        return (feature_size) ** 2 * self.kernel_size[0] ** 2 * \
               self.in_channels * self.out_channels / (self.groups * 2 ** 20)

    def forward(self, input):
        # self.computation = self.get_computation(input)
        self.weight_copy[:] = self.weight.data[:]
        weight_max = \
            torch.abs(self.weight_copy).max(dim=1, keepdim=True)[0].max(
                        dim=2, keepdim=True)[0].max(dim=3, keepdim=True)[0]
        self.weight.data[:] = torch.round((self.weight.data[:] / (weight_max  + 1e-12)) * 127) * weight_max / 127
        out = F.conv2d(input, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)
        self.weight.data[:] = self.weight_copy[:]
        return out

class ConvTranspose2dQuant(nn.ConvTranspose2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, output_padding=0, groups=1, bias=True, dilation=1):
        super(ConvTranspose2dQuant, self).__init__(in_channels, out_channels, kernel_size, stride=stride,
                 padding=padding, output_padding=output_padding, groups=groups, bias=bias, dilation=dilation)
        self.register_buffer('weight_copy', torch.zeros(self.weight.size()))
    def forward(self, input, output_size=None):
        self.weight_copy[:] = self.weight.data[:]
        self.weight.data = self.weight.data.permute(1, 0, 2, 3)
        groups = self.groups
        # cgl -> channelin_group_len
        cgl = int(self.in_channels / groups)
        for i in range(groups):
            subarray = self.weight.data[:,i * cgl: (i + 1) * cgl]
            subarray_max = torch.abs(subarray).max(dim=1, keepdim=True)[0].max(
                                dim=2, keepdim=True)[0].max(dim=3, keepdim=True)[0]
            self.weight.data[:, i * cgl: (i + 1) * cgl] = torch.round(
                (self.weight.data[:, i * cgl: (i + 1) * cgl] / (subarray_max + 1e-12)) * 127) * subarray_max / 127
        self.weight.data = self.weight.data.permute(1, 0, 2, 3)
        out = F.conv_transpose2d(input, self.weight, self.bias, self.stride,
                                 self.padding, self.output_padding, self.groups, self.dilation)
        self.weight.data[:] = self.weight_copy[:]
        return out

class LinearQuant(nn.Linear):
    def __init__(self, in_features, out_features, bias=False):
        super(LinearQuant, self).__init__(in_features, out_features, bias=bias)
        self.register_buffer('weight_copy', torch.zeros(self.weight.size()))

    def forward(self, input):
        self.weight_copy[:] = self.weight.data[:]
        weight_max = \
            torch.abs(self.weight_copy).max(dim=1, keepdim=True)[0]
        self.weight.data[:] = torch.round((self.weight.data[:] / (weight_max  + 1e-12)) * 127) * weight_max / 127
        out = F.linear(input, self.weight, self.bias)
        self.weight.data[:] = self.weight_copy[:]
        return out

class Swish(nn.Module):
    dump_patches = True
    def __init__(self):
        super(Swish, self).__init__()
        self.register_buffer('k', torch.zeros(12))
        self.register_buffer('b', torch.zeros(12))
        self.exp_points()
    def forward(self, x):
        x = F.hardtanh(x, -4, 4)
        x_ = x.unsqueeze(-1)
        x_seg = (-x_)*Variable(self.k, requires_grad=False)+ Variable(self.b, requires_grad=False)
        x_exp, _ = torch.max(x_seg, dim=-1)
        out = x/(1 + x_exp) + 0.2785
        # out = F.hardtanh(out, -4, 4, inplace=False)
        # out.data[:] = torch.round(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out
    def exp_points(self):
        t = np.arange(-6., 7., 1.)
        y_p = 2.**t
        y_p_diff = y_p[1:] - y_p[:-1]
        b = y_p[1:] - y_p_diff*t[1:]
        k = y_p_diff/np.log(2)
        self.k, self.b = (torch.from_numpy(k)).float(), \
                         (torch.from_numpy(b)).float()

# 若训练时采用nn.Sigmoid不近似e指数版本，将导出为type 60(SigmoidAccurate)
# class SigmoidQuant(nn.Module):
#     def __init__(self):
#         super(SigmoidQuant, self).__init__()
#         self.register_buffer('k', torch.zeros(12))
#         self.register_buffer('b', torch.zeros(12))
#         self.exp_points()
#     def forward(self, x):
#         x = F.hardtanh(x, -4, 4)
#         x_ = x.unsqueeze(-1)
#         x_seg = (-x_) * Variable(self.k, requires_grad=False) + Variable(self.b, requires_grad=False)
#         x_exp, _ = torch.max(x_seg, dim=-1)
#         out = 1/(1 + x_exp)
#         out.data[:] = torch.round(out.data[:] * (255. / 1.)) * (1. / 255.)
#         return out
#     def exp_points(self):
#         t = np.arange(-6., 7., 1.)
#         y_p = 2.**t
#         y_p_diff = y_p[1:] - y_p[:-1]
#         b = y_p[1:] - y_p_diff*t[1:]
#         k = y_p_diff/np.log(2)
#         self.k, self.b = (torch.from_numpy(k)).float(), (torch.from_numpy(b)).float()

class SigmoidQuant():
    def __init__(self):
        pass

# 近似e指数，与ycnn底层实现相同
class ExpAprox(nn.Module):
    def __init__(self):
        super(ExpAprox, self).__init__()
    k = torch.zeros(12)
    b = torch.zeros(12)
    t = np.arange(-6., 7., 1.)
    y_p = 2. ** t
    y_p_diff = y_p[1:] - y_p[:-1]
    bb = y_p[1:] - y_p_diff * t[1:]
    kk = y_p_diff / np.log(2)
    k, b = (torch.from_numpy(kk)).float(), (torch.from_numpy(bb)).float()

    def forward(self, x):
        x = F.hardtanh(x, -4, 4)
        x_ = x.unsqueeze(-1)
        x_seg = (x_) * Variable(self.k, requires_grad=False) + Variable(self.b, requires_grad=False)
        x_exp, _ = torch.max(x_seg, dim=-1)
        return x_exp

class NormSigmoid(nn.Module):
    def __init__(self):
        super(NormSigmoid, self).__init__()

    def forward(self, x):
        out = torch.sigmoid(x) / torch.sigmoid(x).sum(dim=1).unsqueeze(1)
        return out

# deprecated: 浮点模块，不带Quant, 直接用AvgPool2d就可以
class AvgPool(nn.AvgPool2d):
    dump_patches = True

    def __init__(self, kernel_size, stride=None, padding=0, ceil_mode=False,
                 count_include_pad=True):
        super(AvgPool, self).__init__(kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode,
                                      count_include_pad=count_include_pad)

    def forward(self, x):
        out = F.avg_pool2d(x, self.kernel_size, self.stride,
                           self.padding, self.ceil_mode, self.count_include_pad)
        # out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class LeakyReLU(nn.Module):
    def __init__(self, val=1e-2):
        super(LeakyReLU, self).__init__()
        self.relu = nn.ReLU()
        self.val = val

    def forward(self, x):
        return x * self.val + self.relu(x) * (1-self.val)

### 一些特殊的Module封装 ###
# set quantize=True if use Quantized model
class RoiAlign(nn.Module):
    dump_patches = True
    def __init__(self, input_size, output_size, min_val=0, max_val=4, quantize=True):
        super(RoiAlign, self).__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.min_val = min_val
        self.max_val = max_val
        self.delta = self.max_val - self.min_val
        self.quantize = quantize

    def forward(self, x):
        input, bbox = x
        bbox = bbox.numpy().astype(float)
        x = torch.linspace(0, self.output_size[0] - 1, self.output_size[0])
        y = torch.linspace(0, self.output_size[1] - 1, self.output_size[1])
        x_grid, y_grid = torch.meshgrid(x, y)
        x_grids = x_grid.unsqueeze(0).repeat(input.size(0), 1, 1).numpy()
        y_grids = y_grid.unsqueeze(0).repeat(input.size(0), 1, 1).numpy()
        x_grids = (((x_grids * (bbox[:, 2] - bbox[:, 0]) + bbox[:, 0]) -0.5) * 2)
        y_grids = (((y_grids * (bbox[:, 3] - bbox[:, 1]) + bbox[:, 1]) -0.5) * 2)
        x_grids = np.expand_dims(x_grids, axis=3)
        y_grids = np.expand_dims(y_grids, axis=3)
        grids = np.concatenate((x_grids, y_grids), axis=3)
        grids = torch.Tensor(grids)
        roi = F.grid_sample(input, grids)
        if self.quantize:
            roi = torch.round(roi * (255. / self.delta)) * (self.delta / 255.)
        return roi

class H_Sigmoid(nn.Module):
    def __init__(self):
        super(H_Sigmoid, self).__init__()

    def forward(self, x):
        out = F.hardtanh(x + 3, 0, 6) / 6
        return out

# 此op，后面不可加cat，add，avg
class H_SigmoidQuant(nn.Module):
    def __init__(self, min_val=0, max_val=1, inplace=False):
        super(H_SigmoidQuant, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.delta = self.max_val - self.min_val
        self.inplace = inplace

    def forward(self, x):
        out = F.hardtanh(x + 3, 0, 6) / 6
        if self.min_val == 0:  # 只有最小为0时，需要量化，否则输出是浮点
            out.data[:] = torch.round(out.data[:] * (255. / self.delta)) * (self.delta / 255.)
        else:
            raise Exception("只有最小为0时，需要量化，否则输出是浮点")
        return out

class H_Swish(nn.Module):
    def __init__(self):
        super(H_Swish, self).__init__()

    def forward(self, x):
        out = x * F.hardtanh(x + 3, 0, 6) / 6
        return out

class H_SwishQuant(nn.Module):
    def __init__(self, min_val = 0, max_val = 4, inplace = False):
        super(H_SwishQuant, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.delta = self.max_val - self.min_val
        self.inplace = inplace

    def forward(self, x):
        out = F.hardtanh(x, self.min_val, self.max_val, self.inplace)
        out = out * F.hardtanh(out + 3, 0, 6) / 6
        if self.min_val == 0:  # 只有最小为0时，需要量化，否则输出是浮点
            out.data[:] = torch.round(out.data[:] * (255. / self.delta)) * (self.delta / 255.)
        else:
            raise Exception("只有最小为0时，需要量化，否则输出是浮点")
        return out


class ILN(nn.Module):
    def __init__(self, num_features, eps=1e-5, unbiased=True):
        super(ILN, self).__init__()
        self.eps = eps
        self.unbiased = unbiased
        self.num_features = num_features
        self.rho = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.gamma = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.beta = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.rho.data.fill_(0.0)
        self.gamma.data.fill_(1.0)
        self.beta.data.fill_(0.0)

    def forward(self, input):
        # in_mean, in_var = torch.mean(torch.mean(input, dim=2, keepdim=True), dim=3, keepdim=True),
        # torch.var(torch.var(input, dim=2, keepdim=True), dim=3, keepdim=True)
        # in_mean, in_var = torch.mean(input, dim=[2,3], keepdim=True),
        # torch.var(input, dim=[2,3], keepdim=True)
        n, c, h, w = input.shape
        input_in = input.view(n, c, -1)
        in_mean, in_var = torch.mean(input_in, dim=2, keepdim=True).unsqueeze(-1), \
                          torch.var(input_in, dim=2, keepdim=True, unbiased=self.unbiased).unsqueeze(-1)
        out_in = (input - in_mean) / torch.sqrt(in_var + self.eps)

        input_ln = input.view(n, -1)
        ln_mean, ln_var = torch.mean(input_ln, dim=1, keepdim=True).unsqueeze(-1).unsqueeze(-1), \
                          torch.var(input_ln, dim=1, keepdim=True, unbiased=self.unbiased).unsqueeze(-1).unsqueeze(-1)
        out_ln = (input - ln_mean) / torch.sqrt(ln_var + self.eps)

        # self.rho.data = self.rho.data.clamp(0.0, 1.0)
        self.rho = self.rho.clamp(0.0, 1.0)

        out = self.rho.expand(input.shape[0], -1, -1, -1) * out_in + \
                    (1-self.rho.expand(input.shape[0], -1, -1, -1)) * out_ln
        out = out * self.gamma.expand(input.shape[0], -1, -1, -1) + \
                    self.beta.expand(input.shape[0], -1, -1, -1)
        return out

class WeightCenter(nn.Module):
    dump_patches = True
    def __init__(self, height, width):
        super(WeightCenter, self).__init__()
        self.height = height
        self.width = width
        pos_x, pos_y = np.meshgrid(
            np.linspace(0., 1., self.height),
            np.linspace(0., 1., self.width)
        )
        # TODO: reshape of not according to .view
        pos_x = torch.from_numpy(pos_x.reshape(self.height * self.width)).float()
        pos_y = torch.from_numpy(pos_y.reshape(self.height * self.width)).float()
        self.register_buffer('pos_x', pos_x)
        self.register_buffer('pos_y', pos_y)

    def forward(self, x):
        expected_x = torch.sum(Variable(self.pos_x, requires_grad=False) * x, dim=2)
        expected_y = torch.sum(Variable(self.pos_y, requires_grad=False) * x, dim=2)
        expected_x_numpy = expected_x.cpu().numpy()
        expected_y_numpy = expected_y.cpu().numpy()
        expected_xy_numpy = np.concatenate((expected_x_numpy, expected_y_numpy), axis=1)
        expected_xy = torch.from_numpy(expected_xy_numpy)
        if torch.cuda.is_available():
            expected_xy = expected_xy.cuda()
        expected_xy = Variable(expected_xy, requires_grad=False)
        return expected_xy

class Reorg(nn.Module):
    dump_patches = True

    def __init__(self):
        super(Reorg, self).__init__()

    def forward(self, x):
        ss = x.size()
        out = x.view(ss[0], ss[1], ss[2] // 2, 2, ss[3]).view(ss[0], ss[1], ss[2] // 2, 2, ss[3] // 2, 2).\
            permute(0, 1, 3, 5, 2, 4).contiguous().view(ss[0], -1, ss[2] // 2, ss[3] // 2)
        return out

class ShuffleBlock(nn.Module):
    def __init__(self, groups):
        super(ShuffleBlock, self).__init__()
        self.groups = groups

    def forward(self, x):
        '''Channel shuffle: [N,C,H,W] -> [N,g,C/g,H,W] -> [N,C/g,g,H,w] -> [N,C,H,W]'''
        N, C, H, W = x.size()
        g = self.groups
        return x.view(N, g, C // g, H, W).permute(0, 2, 1, 3, 4).reshape(N, C, H, W)

class DynamicConv2d(nn.Conv2d):

    dump_patches = True
    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=1, padding=0, dilation=1, groups=1, bias=False):
        super(DynamicConv2d, self).__init__(in_channels, out_channels,
                                          kernel_size, stride, padding,
                                          dilation, groups=groups, bias=bias)
        self.computation = 0

    def forward(self, inputs):
        input, weight = inputs[0], inputs[1]

        if isinstance(self.kernel_size, tuple) or isinstance(self.kernel_size, list):
            weight = weight.view(self.out_channels, self.in_channels,
                                 self.kernel_size[0], self.kernel_size[1])
        else:
            weight = weight.view(self.out_channels, self.in_channels,
                                 self.kernel_size, self.kernel_size)
        out = F.conv2d(input, weight, self.bias,
                       self.stride, self.padding, self.dilation, self.groups)
        return out

class DynamicConvTranspose2d(nn.ConvTranspose2d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, output_padding=0, groups=1, bias=True, dilation=1):
        super(DynamicConvTranspose2d, self).__init__(in_channels, out_channels, kernel_size, stride=stride,
                 padding=padding, output_padding=output_padding, groups=groups, bias=bias, dilation=dilation)

    def forward(self, inputs, output_size=None):
        input, weight = inputs[0], inputs[1]
        out = F.conv_transpose2d(input, weight, self.bias, self.stride,
                                 self.padding, self.output_padding, self.groups, self.dilation)
        self.weight.data[:] = self.weight_copy[:]
        return out

#========================================================================================================
def conv_layer(channel_in, channel_out, ks=1, stride=1, padding=0, dilation=1,
               bias=False, bn=True, cut=True, relu=True,
               group=1, weight_quant=True):
    if weight_quant:
        _conv = Conv2dQuant
    else:
        _conv = nn.Conv2d
    sequence = [_conv(channel_in, channel_out,
                      kernel_size=ks, stride=stride, padding=padding,
                      dilation=dilation, bias=bias, groups=group)]
    if bn:
        sequence.append(nn.BatchNorm2d(channel_out))
    if relu:
        if cut:
            if not weight_quant:
                print ('error: float must: weight_quant({}) -> false, cut({}) -> false(if relu is true).'.format(
                        weight_quant, relu, cut))
                raise
            sequence.append(HardQuant(0, 4))
        else:
            # if not weight_quant:
            #     print ('error: fix must: weight_quant({}) -> true, relu({}) -> true, cut({}) -> true.'.format(
            #             weight_quant, relu, cut))
            #     raise
            sequence.append(nn.ReLU())
    else:
        if cut:  # TODO
            pass#sequence.append(HardQuant(0, 4))
        else:
            print("Warning! No activation layer followed after Conv!")
    print('computation:', channel_in, channel_out, ks, stride)
    return nn.Sequential(*sequence)

def linear_layer(channel_in, channel_out, bias=False, bn=True, relu=True, unit=False, weight_quant=True):
    if not unit:
        if weight_quant:
            _linear = LinearQuant
        else:
            _linear = nn.Linear
        sequence = [_linear(channel_in, channel_out, bias=bias)]
    else:
        sequence = [LinearUnit(channel_in, channel_out, bias=bias)]
    if bn:
        sequence.append(nn.BatchNorm1d(channel_out))
    if relu:
        sequence.append(HardQuant(0, 4))

    # if weight_quant and (not relu):
    #     print('error: fix must: weight_quant({}) -> true, relu({}) -> true.'.format(
    #         weight_quant, relu))
    #     raise

    return nn.Sequential(*sequence)

class LinearUnit(nn.Linear):
    def __init__(self, in_features, out_features, bias=True):
        super(LinearUnit, self).__init__(in_features, out_features, bias=bias)
        self.weight_unit = None
    def forward(self, input):
        weight_norm = torch.sqrt((self.weight ** 2).sum(dim=1, keepdim=True))
        self.weight_unit = self.weight / weight_norm
        out = F.linear(input, self.weight_unit, self.bias)
        return out

class mobile_unit(nn.Module):
    dump_patches = True

    def __init__(self, channel_in, channel_out, stride=1, num3x3=1, has_half_out=False):
        print('Making mobile_unit!')
        super(mobile_unit, self).__init__()
        self.stride = stride
        self.channel_in = channel_in
        self.channel_out = channel_out
        if num3x3 == 1:
            self.conv3x3 = nn.Sequential(
                conv_layer(channel_in, channel_in, ks=3, stride=stride, padding=1, group=channel_in),
            )
        else:
            self.conv3x3 = nn.Sequential(
                conv_layer(channel_in, channel_in, ks=3, stride=1, padding=1, group=channel_in),
                conv_layer(channel_in, channel_in, ks=3, stride=stride, padding=1, group=channel_in),
            )
        self.conv1x1 = conv_layer(channel_in, channel_out)
        self.has_half_out = has_half_out
        self.avg = AvgQuant()

    def forward(self, x):
        half_out = self.conv3x3(x)
        out = self.conv1x1(half_out)
        if self.stride == 1 and (self.channel_in == self.channel_out):
            out = self.avg(out, x)
        if self.has_half_out:
            return half_out, out
        else:
            return out

class mobile_unit_v2(nn.Module):
    dump_patches = True
    def __init__(self, channel_in, channel_out, stride=1, num3x3=1):
        print('Making mobile_unit_v2!')
        super(mobile_unit_v2, self).__init__()
        self.stride = stride
        self.channel_in = channel_in
        self.channel_out = channel_out
        if self.stride > 1 or self.channel_in < self.channel_out:
            channel_3x3 = channel_in
        else:
            channel_3x3 = channel_out

        if num3x3 == 1:
            self.conv3x3 = nn.Sequential(
                conv_layer(channel_3x3, channel_3x3, ks=3, stride=stride, padding=1, group=channel_3x3),
            )
        else:
            self.conv3x3 = nn.Sequential(
                conv_layer(channel_3x3, channel_3x3, ks=3, stride=1, padding=1, group=channel_3x3),
                conv_layer(channel_3x3, channel_3x3, ks=3, stride=stride, padding=1, group=channel_3x3),
            )
        self.conv1x1 = conv_layer(channel_in, channel_out)
        self.avg = AvgQuant()

    def forward(self, x):
        if self.stride > 1 or self.channel_in < self.channel_out:
            half_out = self.conv3x3(x)
            out = self.conv1x1(half_out)
        else:
            out = self.conv1x1(x)
            out = self.conv3x3(out)
        if self.stride == 1 and (self.channel_in == self.channel_out):
            out = self.avg(out, x)
        return out

class DenseBlock(nn.Module):
    dump_patches = True

    def __init__(self, channel_in):
        super(DenseBlock, self).__init__()
        self.channel_in = channel_in
        self.conv1 = mobile_unit_v2(channel_in, channel_in//4)
        self.conv2 = mobile_unit_v2(channel_in*5//4, channel_in//4)
        self.conv3 = mobile_unit_v2(channel_in*6//4, channel_in//4)
        self.conv4 = mobile_unit_v2(channel_in*7//4, channel_in//4)

    def forward(self, x):
        out1 = self.conv1(x)
        comb1 = torch.cat([x, out1])
        out2 = self.conv2(comb1)
        comb2 = torch.cat([comb1, out2])
        out3 = self.conv3(comb2)
        comb3 = torch.cat([comb2, out3])
        out4 = self.conv4(comb3)
        comb4 = torch.cat([comb3, out4])
        return comb4

class DenseBlock2(nn.Module):
    dump_patches = True

    def __init__(self, channel_in):
        super(DenseBlock2, self).__init__()
        self.channel_in = channel_in
        self.conv1 = mobile_unit_v2(channel_in, channel_in//2)
        self.conv2 = mobile_unit_v2(channel_in*3//2, channel_in//2)

    def forward(self, x):
        out1 = self.conv1(x)
        comb1 = torch.cat([x, out1])
        out2 = self.conv2(comb1)
        comb2 = torch.cat([comb1, out2])
        return comb2


### 以下class为导模型内部使用标记，不作为训练使用 ###
class SliceAdd():
    def __init__(self):
        pass

class ElementwiseAdd():
    def __init__(self):
        pass

class ElementwiseSub():
    def __int__(self):
        pass

class MulConstant():
    def __init__(self):
        pass

class Reshape():
    def __init__(self):
        pass

class View():
    def __init__(self):
        pass

class Mul():
    def __init__(self):
        pass

class SubConstant():
    def __init__(self):
        pass

class Concat():
    def __init__(self):
        pass

class Transpose():
    def __init__(self):
        pass

class BMM():
    def __init__(self):
        pass

class Sum():
    def __init__(self):
        pass

class SliceDiv():
    def __init__(self):
        pass

class SlicePickRange():
    def __init__(self):
        pass

class DirectCopy():
    def __init__(self, index):
        self.index = index
        setattr(self, '__name__', 'DirectCopy')

    def __str__(self):
        return 'DirectCopy'

class Scatter():
    def __init__(self):
        pass

class Permute():
    def __init__(self):
        pass

class Clone():
    def __init__(self):
        pass

class PixelMul():
    def __init__(self):
        pass

class Exp():
    def __init__(self):
        pass

class Unique():
    def __init__(self):
        pass

class Pick():
    def __init__(self):
        pass

class Equal():
    def __init__(self):
        pass

class Pow():
    def __init__(self):
        pass

class ExpandAs():
    def __init__(self):
        pass


class Repeat():
    def __init__(self):
        pass

class Inverse():
    def __init__(self):
        pass

class DirectNormalCopy():
    def __init__(self, index):
        self.index = index
        setattr(self, '__name__', 'DirectNormalCopy')

    def __str__(self):
        return 'DirectNormalCopy'

class Abs():
    def __init__(self):
        pass

# 该模块是为了兼容之前的旧模型
# 模型中使用torch.cat()，不要使用本Cat
class Cat(nn.Module):
    dump_patches = True
    def __init__(self, dim=1):
        super(Cat, self).__init__()
        self.__name__ = "Cat"
        self.dim = dim

    def forward(self, x):
         return torch.cat([x[0], x[1]], dim=self.dim)
    # For RNN model
    #def forward(self, x1, x2):
    #    return torch.cat([x1, x2], dim=1)

class AdaptiveAvgPool(nn.AdaptiveAvgPool2d):
    dump_patches = True
    def __init__(self, output_size):
        super(AdaptiveAvgPool, self).__init__(output_size)
    def forward(self, x):
        out = F.adaptive_avg_pool2d(x, self.output_size)
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class RCConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(RCConv2d, self).__init__(in_channels, out_channels, kernel_size,
                                       stride, padding, dilation, groups, bias)
        self.enable_mac_cnt = False
        self.mac_cnt = None
        self.mac_cnt_bias = None
        # in_bn, out_bn layers, None means to be filled, 1 means to be copied (skip connection)
        self.in_out_bn = [None, None]
        self._wn = False
        self.prev_conv = []

    @property
    def wn(self):
        return self._wn

    @wn.setter
    def wn(self, val):
        if self.in_out_bn[0] is None:
            raise ValueError
        in_bn = self.in_out_bn[0]
        if isinstance(in_bn, list):
            self.id_in_channels = sum([bn.weight.numel() for bn in in_bn])
        else:
            self.id_in_channels = in_bn.weight.numel()
        assert self.id_in_channels == self.in_channels or self.id_in_channels * 4 == self.in_channels
        if self.groups == 1:
            channel_norms = self.weight.data.transpose(0, 1).contiguous().view(
                    self.id_in_channels, -1).norm(p=2, dim=-1).clamp_(min=1e-12)
        else:
            channel_norms = self.weight.data.view(self.id_in_channels, -1).norm(p=2, dim=-1).clamp_(min=1e-12)
        if val:
            if isinstance(in_bn, list):
                i = 0
                for bn in in_bn:
                    bn.weight.data *= channel_norms[i:i+bn.weight.numel()]
                    bn.bias.data *= channel_norms[i:i+bn.bias.numel()]
                    i += bn.weight.numel()
            else:
                in_bn.weight.data *= channel_norms
                in_bn.bias.data *= channel_norms
        else:
            self.weight.data /= channel_norms.view(1, -1, 1, 1)
        self._wn = val

    def set_in_bn(self, bn_layer):
        self.in_out_bn[0] = bn_layer

    def get_in_bn(self):
        return self.in_out_bn[0]

    def set_out_bn(self, bn_layer):
        self.in_out_bn[1] = bn_layer

    def get_out_bn(self):
        return self.in_out_bn[1]

    def set_pre_conv(self, prev_conv_layers):
        self.prev_conv = prev_conv_layers

    def forward(self, input):
        if self.enable_mac_cnt:
            fake_output = F.conv2d(torch.ones_like(input.data), torch.ones_like(self.weight.data), None,
                                   self.stride, self.padding, self.dilation, self.groups)
            self.mac_cnt_bias = float(fake_output.numel())
            self.mac_cnt = fake_output.sum().item()
        if self.wn:
            if self.groups == 1:
                channel_norms = self.weight.data.transpose(0, 1).view(
                     self.id_in_channels, -1).norm(p=2, dim=-1).clamp_(min=1e-12)
            else:
                channel_norms = self.weight.data.view(self.id_in_channels, -1).norm(p=2, dim=-1).clamp_(min=1e-12)
            weight = (self.weight.view(self.weight.size(0), channel_norms.numel(), -1) /
                      channel_norms.view(1, -1, 1)).view(self.weight.shape)
        else:
            weight = self.weight
        return F.conv2d(input, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)

#============================== for new op

class AdaptiveAvg(nn.Module):
    dump_patches = True
    def __init__(self, weights=None):
        super(AdaptiveAvg, self).__init__()
        self.weights = weights
    def forward(self, x):
        if isinstance(x, list) or isinstance(x, tuple):
            if self.weights is None:
                self.weights = [1.0/len(list(x)) for i in range(len(list(x)))]
            out = x[0] * self.weights[0]
            for i in range(1, len(x)):
                out += x[i] * self.weights[i]
        else:
            raise ValueError('the type of weights is error.')
        return out

class AdaptiveAvgQuant(nn.Module):
    dump_patches = True
    def __init__(self, weights=None):
        super(AdaptiveAvgQuant, self).__init__()
        self.weights = weights
    def forward(self, x):
        if isinstance(x, list) or isinstance(x, tuple):
            if self.weights is None:
                self.weights = [1.0/len(list(x)) for i in range(len(list(x)))]
            out = x[0] * self.weights[0]
            for i in range(1, len(x)):
                out += x[i] * self.weights[i]
        else:
            raise ValueError('the inputs of AdaptiveAvgQuant must be list/tuple.')
        out.data[:] = torch.floor(out.data[:] * (255. / 4.)) * (4. / 255.)
        return out

class SigmoidOp(nn.Module):
    dump_patches = True
    def __init__(self, data_in_type="float", data_out_type="float"):
        super(SigmoidOp, self).__init__()
        self.data_in_type = data_in_type # 'float/fixed/None'
        self.data_out_type = data_out_type  # 'float/fixed/None'

    def set_data_in_type(self, data_in_type):
        self.data_in_type = data_in_type

    def set_data_out_type(self, data_out_type):
        self.data_out_type = data_out_type

    def forward(self, x):
        out = torch.sigmoid(x)
        if self.data_out_type == "fixed":
            out.data[:] = torch.round(out.data[:] * (255. / 1.)) * (1. / 255.)
        return out

class TanhOp(nn.Module):
    dump_patches = True

    def __init__(self, data_in_type=None, data_out_type=None):
        super(TanhOp, self).__init__()
        self.data_in_type = data_in_type  # 'float/fixed/None'
        self.data_out_type = data_out_type  # 'float/fixed/None'

    def set_data_in_type(self, data_in_type):
        self.data_in_type = data_in_type

    def set_data_out_type(self, data_out_type):
        self.data_out_type = data_out_type

    def forward(self, x):
        out = torch.tanh(x)
        return out

class SoftMaxOp(nn.Module):
    dump_patches = True

    def __init__(self, data_in_type='float', data_out_type='float', dim=1):
        super(SoftMaxOp, self).__init__()
        self.data_in_type = data_in_type  # 'float/fixed/None'
        self.data_out_type = data_out_type  # 'float/fixed/None'
        self.dim = dim

    def set_data_in_type(self, data_in_type):
        self.data_in_type = data_in_type

    def set_data_out_type(self, data_out_type):
        self.data_out_type = data_out_type

    def forward(self, inputs):
        out = torch.softmax(inputs, dim=self.dim)
        return out

class AdaptivePad(nn.Module):
    def __init__(self, pad, mode='constant', value=0):
        super(AdaptivePad, self).__init__()

        if not(isinstance(pad, list) or isinstance(pad, tuple)):
            raise ValueError("AdaptivePad: pad must be list or tuple.")
        if len(pad) != 6:
            raise ValueError("AdaptivePad: len(pad) must be equal to 6.")

        if mode not in ["constant", "reflect"]:
            raise ValueError("AdaptivePad: mode must be in [constant, reflect].")

        self.pad = pad
        self.mode = mode
        self.value = value

    def forward(self, x):
        out = F.pad(x, self.pad, mode=self.mode, value=self.value)
        return out

class WingLoss(nn.Module):
    def __init__(self, omega, sigma):
        super(WingLoss, self).__init__()
        self.omega = omega
        self.sigma = sigma
        assert omega > 0, 'omage must be positive'
        assert sigma > 0, 'sigma must be positive'
        self.c = omega - omega*np.log(1+omega/sigma)

    def forward(self, x, t):
        diff_abs = torch.abs(x - t)
        small_loss = self.omega*torch.log(1+diff_abs/self.sigma)
        big_loss = diff_abs - self.c
        is_small = (diff_abs < self.omega).float()
        return torch.mean(small_loss*is_small + big_loss*(1-is_small))


class AlphaFusion(nn.Module):
    def __init__(self, data_type="float"): # data_type = "float/fixed"
        super(AlphaFusion, self).__init__()
        self.data_in_type = data_type
        self.data_out_type = data_type

    def forward(self, f1, f2, alpha):
        out = alpha * f1 + (1 - alpha) * f2
        if self.data_out_type == "fixed":
            out.data[:] = torch.floor(out.data[:] * (255. / 256.) * (255. / 4.)) * (4. / 255.)
        return out

class Filter(nn.Module):
    def __init__(self, threshold=None):
        super(Filter, self).__init__()
        self.th = threshold

    def forward(self, x):
        forward_config.reset_old_forward()
        if len(x) == 1:
            out = x[x > self.threshold]
        else:
            obj, score = x
            score = score.squeeze(0)
            obj = obj.squeeze(0)
            out = obj[score > self.th] if self.th is not None else obj[score]
        out = out.unsqueeze(0)
        forward_config.reset_new_forward()
        return out


class IndexSelect(nn.Module):
    def __init__(self, dim):
        super(IndexSelect, self).__init__()
        self.dim = dim

    def forward(self, inputs):
        x, col = inputs
        col = col.to(torch.int64)
        out = torch.index_select(x, self.dim, col.squeeze(0))
        return out

class Size(nn.Module):
    def __init__(self, dim):
        super(Size, self).__init__()
        self.dim = dim

    def forward(self, x):
        out = torch.ones(1) * x.size(self.dim)
        out = out.unsqueeze(0)
        return out


class ZerosOrOnes(nn.Module):

    def __init__(self, mode):
        super(ZerosOrOnes, self).__init__()

        self.mode = mode  # 0, zeros; 1, ones;

    # [tensor, ]
    def forward(self, x):
        shape = [1]

        if not (isinstance(x, list) or isinstance(x, tuple)):
            raise ValueError("ZerosOrOnes: inputs must be list or tuple")

        for idx in x:
            if len(idx.size()) != 2:
                raise ValueError("ZerosOrOnes: dims of inputs must be 2")
            shape += [int(idx.numpy()[0, 0])]

        if self.mode == 0:
            return torch.zeros(shape)
        elif self.mode == 1:
            return torch.ones(shape)
        else:
            raise ValueError("ZerosOrOnes: mode is error！")


class ScatterAdd(nn.Module):
    def __init__(self, dim) -> None:
        super(ScatterAdd, self).__init__()
        self.dim = dim

    def forward(self, inputs):
        x, y, idx = inputs
        idx = idx.to(torch.int64)
        out = torch.scatter_add(x, self.dim, idx, y)
        return out

class BatchConv(nn.Module):

    '''
    :inputs
        - x : tensor, (n, c, h, w) -> (1, c, batch, h*w)
    :param
        - input_shape : (batch, c, h, w), feature shape for conv or dw_conv
        - conv_mode : 'conv' or 'dw_conv'
        - ks: kernel size, [kh, kw]
    '''

    def __init__(self,
                 input_shape: list, kernel_size: list,
                 in_channels: int, out_channels: int,
                 stride: int = 1, padding: int = 0, dilation: int = 1,
                 bias: bool = False, conv_mode: str = 'conv'):

        super(BatchConv, self).__init__()

        assert conv_mode in ['conv', 'dw_conv']

        self.input_shape = input_shape
        self.mode = conv_mode
        self.ks = kernel_size
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.bias = bias

        self.output_shape = list()

        if self.mode == 'dw_conv':
            self.spatial_layer = nn.Conv2d(self.in_channels, self.out_channels,
                                           kernel_size=self.ks, stride=self.stride,
                                           padding=self.padding, dilation=self.dilation,
                                           groups=self.in_channels, bias=False)
        else:
            self.spatial_layer = nn.Conv2d(self.in_channels, self.out_channels,
                                           kernel_size=self.ks, stride=self.stride,
                                           padding=self.padding, bias=False)

    def forward(self, x):
        forward_config.reset_old_forward()

        # (1, c, batch, h*w)
        # (batch, c, h*w, 1)
        x = x.permute(2, 1, 3, 0)
        # (batch, c, h, w)
        x = x.view(self.input_shape[0], self.input_shape[1],
                   self.input_shape[2], self.input_shape[3])
        # (batch, c_out, h_out, w_out)
        x = self.spatial_layer(x)
        self.output_shape.append(list(x.size()))
        # (1, c_out, batch, h*w)
        x = x.view(1, self.input_shape[0], self.out_channels, -1)
        x = x.permute(0, 2, 1, 3)

        forward_config.reset_new_forward()
        return x

class SoftMaxAccurateOp(nn.Module):
    dump_patches = True

    def __init__(self, data_in_type='float', data_out_type='float', dim=1):
        super(SoftMaxAccurateOp, self).__init__()
        self.data_in_type = data_in_type  # 'float/fixed/None'
        self.data_out_type = data_out_type  # 'float/fixed/None'
        self.dim = dim

    def set_data_in_type(self, data_in_type):
        self.data_in_type = data_in_type

    def set_data_out_type(self, data_out_type):
        self.data_out_type = data_out_type

    def forward(self, inputs):
        out = torch.softmax(inputs, dim=self.dim)
        return out

class GridSample2D(nn.Module):
    def __init__(self,
        mode='bilinear',
        padding_mode='border',
        align_corners=True,
        quantize=False,
        min_val=0.0,
        max_val=4.0
    ):
        super(GridSample2D, self).__init__()
        self.mode = mode
        self.padding_mode = padding_mode
        self.align_corners = align_corners
        self.quantize = quantize
        self.min_val = min_val
        self.max_val = max_val

    def forward(self, x):
        input, grid = x
        out = F.grid_sample(
            input, grid,
            mode=self.mode,
            padding_mode=self.padding_mode,
            align_corners=self.align_corners
        )
        if self.quantize:
            delta = self.max_val - self.min_val
            out = torch.round(out * (255. / delta)) * (delta / 255.)
        return out

class MatMul(nn.Module):
    def __init__(self):
        super(MatMul, self).__init__()

        self.input_shape = []

    def isSupported(self, x, y):
        size_x = x.size()
        size_y = y.size()

        assert len(size_x) != 1 or len(size_y) != 1
        assert size_x[0] == 1 or size_y[0] == 1

        self.input_shape.append([size_x])


    def forward(self, x, y):
       self.isSupported(x, y)
       return torch.matmul(x, y)

#==========================================================================
def wing_loss(x, t, omega, sigma):
    c = omega - omega * np.log(1 + omega / sigma)
    diff_abs = torch.abs(x - t)
    small_loss = omega * torch.log(1 + diff_abs / sigma)
    big_loss = diff_abs - c
    is_small = (diff_abs < omega).float()
    return torch.mean(small_loss * is_small + big_loss * (1 - is_small))

class mobile_unit_for_large_model(nn.Module):
    dump_patches = True

    def __init__(self, channel_in, channel_out, stride=1):
        print('Making PoseModel!')
        super(mobile_unit_for_large_model, self).__init__()
        self.stride = stride
        self.channel_in = channel_in
        self.channel_out = channel_out
        if self.stride > 1 or self.channel_in < self.channel_out:
            channel_3x3 = channel_in
        else:
            channel_3x3 = channel_out

        if stride == 2:
            self.conv3x3 = nn.Sequential(
                AvgPool(kernel_size=stride, stride=stride)
            )
        else:
            self.conv3x3 = nn.Sequential(
                conv_layer(channel_3x3, channel_3x3, ks=3, stride=stride, padding=1, group=channel_3x3),
            )
        self.conv1x1 = conv_layer(channel_in, channel_out)
        self.avg = AvgQuant()

    def forward(self, x):
        if self.stride > 1 or self.channel_in < self.channel_out:
            half_out = self.conv3x3(x)
            out = self.conv1x1(half_out)
        else:
            out = self.conv1x1(x)
            out = self.conv3x3(out)
        if self.stride == 1 and (self.channel_in == self.channel_out):
            out = self.avg(out, x)
        return out

class Clamp(nn.Module):
    dump_patches = True

    def __init__(self, min_val=0, max_val=4):
        self.min_val = min_val
        self.max_val = max_val

    def forward(self, x):
        return torch.clamp(x, self.min_val, self.max_val)


#==================================================================================

LAYER_TYPE_NUM = {
    # Conv2D/Linear
    nn.Conv2d: 38, Conv2dQuant: 9, nn.ConvTranspose2d: 5, nn.Conv1d: 51,
    ConvTranspose2dQuant: 31, LinearQuant: 30, nn.Linear: 25, Assign_Conv2D : 39,
    # Norm
    nn.BatchNorm2d: 8, nn.BatchNorm1d: 8, nn.InstanceNorm2d: 56, ILN: 288,
    nn.LayerNorm: 88,
    # Activation
    HardQuant: 40, QuantizedHardtanh: 40, #SigmoidQuant: 16,
    H_Swish: 70, H_SwishQuant: 70, H_Sigmoid: 71, H_SigmoidQuant: 71,
    nn.Hardtanh: 40, nn.ReLU: 21, nn.LeakyReLU:69, nn.LogSoftmax: 34,
    SigmoidOp: 60, nn.Sigmoid: 60, TanhOp: 55, nn.Tanh: 55,
    SoftMaxOp: 34, nn.Softmax: 34,
    # Pool
    nn.MaxPool2d: 7, nn.MaxPool1d: 52,
    nn.AvgPool2d: 15, AvgPoolQuant: 15, AvgPool: 15,
    AdaptiveAvgPoolQuant: 41, nn.AdaptiveAvgPool2d: 41,
    # Pad
    nn.ConstantPad2d: 87, nn.ReflectionPad2d: 87,
    # Pixel Op
    SliceAdd: 18, SliceAddQuant: 18,
    ElementwiseAdd: 18, ElementwiseSub: 43, Sum: 278, # AddQuant: 18,
    MulConstant: 11, SliceMul: 261, Mul: 19, SubConstant: 45,
    SliceDiv: 279, AvgQuant: 29, BMM: 61, AdaptiveAvgQuant: 289,
    nn.PixelShuffle: 67, SliceMulParams: 269, MulQuant: 19,
    # Mixed Op
    View: 65, Reshape: 26, Concat: 12, Cat: 12, nn.Upsample: 13, UpsampleQuant: 13,
    # RNN
    nn.LSTM: 58, nn.GRU: 49, nn.GRUCell: 50, nn.LSTMCell: 57,
    # Other Old Op
    RoiAlign : 72, RCConv2d: 38, Reorg: 28, Transpose: -1, SlicePickRange: 270,
    WeightCenter: 280, ShuffleBlock: 287, nn.Embedding: 62,
    #new op
    Scatter: 307, Permute: 275, TopK: 308, Coords: 309, PointSample: 310,
    DirectCopy: 44, nn.ELU: 314, Clone: 313, PixelMul: 262, Swish: 22,
    BboxConvert: 311, ROIAlign: 312, AdaptivePad: 316, AlphaFusion: 317,
    DynamicConv2d: 319, DynamicConvTranspose2d: 320,
    MaxOp: 400, MeanOp: 401, NormOp: 402, Exp: 403, IndexSelect: 404, ScatterAdd: 405,
    nn.Unfold: 406, CompareOp: 407, ZerosOrOnes: 408, Size: 409, ExpandAs: 410,
    BatchConv: 411, SoftMaxAccurateOp: 412, Inverse: 414, MinOp: 413,
    GridSample2D: 415, Repeat: 416, Split: 417, DirectNormalCopy: 418, torch.pow: 419,
    Pow: 419, MatMul: 420, Abs: 421,
    # mask_rcnn
    fpn_rpn_outputs: 301, roi_2mlp_head: 302, fast_rcnn_outputs: 303,
    RoiSelection: 304, mask_rcnn_outputs: 305, mask_rcnn_fcn_head_v1upXconvs: 306,
}

if USE_WIREFRAME:
    LAYER_TYPE_NUM.update({
        ProposalLines: 318, LineJuncPick: 319, LineJuncDecoder: 320, LOIPooling: 321,
    })

SKIP_LAYERS = [Transpose]
### 以上为导模型内部使用标记，不作为训练使用 ###
PARAM_LAYERTYPES = [
                    nn.Conv2d, Conv2dQuant, nn.BatchNorm2d, nn.BatchNorm1d,
                    nn.Linear, LinearQuant, Reshape, SlicePickRange,
                    nn.ConvTranspose2d, nn.LSTM, ConvTranspose2dQuant,
                    nn.InstanceNorm2d, RCConv2d, ILN, Assign_Conv2D,
                    nn.LayerNorm, DynamicConvTranspose2d, DynamicConv2d,
                ]

CONV_LAYERS = [
    nn.Conv2d, nn.Linear, LinearQuant, Conv2dQuant,
    nn.Conv1d, nn.ConvTranspose2d, ConvTranspose2dQuant,
    RCConv2d, Assign_Conv2D, DynamicConvTranspose2d,
    DynamicConv2d
]


FUNC_LAYERS = []

# ================== for new framework
MASK_RCNN = {# RPN
             'rpn': [fpn_rpn_outputs],
             'box_head': [roi_2mlp_head],
             'box_outs': [fast_rcnn_outputs],
             'roi_selection': [RoiSelection],
             'mask_head': [mask_rcnn_fcn_head_v1upXconvs],
             'mask_outs': [mask_rcnn_outputs]
            }

WIREFRAME = {
    'proposal_lines': [ProposalLines],
    'lines_junc_pick': [LineJuncPick],
    'lines_junc_decoder': [LineJuncDecoder],
    'loi_pooing': [LOIPooling]
}

RPN_MODULE_NAME = {fpn_rpn_outputs: 'fpn_rpn_outputs'}
BOX_HEAD_MODULE_NAME = {roi_2mlp_head: 'roi_2mlp_head'}
BOX_OUTS = {fast_rcnn_outputs: 'fast_rcnn_outputs'}
ROI_SELECT = {RoiSelection: 'RoiSelection'}
MASK_HEAD = {mask_rcnn_fcn_head_v1upXconvs: 'mask_rcnn_fcn_head_v1upXconvs'}
MASK_OUTS = {mask_rcnn_outputs: 'mask_rcnn_outputs'}

OUTPUT_SIZE_INFO = [
    # mask_rcnn
    fpn_rpn_outputs, DirectCopy, roi_2mlp_head,
    mask_rcnn_fcn_head_v1upXconvs, mask_rcnn_outputs,
    # wireframe
    LineJuncPick, LineJuncDecoder, LOIPooling,
    # alone op
    Split, DirectNormalCopy
]

# ================== for forward

ONE_LAYERTYPES = [
    # old op
    nn.Conv2d, Conv2dQuant, nn.Softmax, nn.ConvTranspose2d, nn.MaxPool2d, nn.BatchNorm2d, nn.BatchNorm1d,
    nn.ReLU, nn.AdaptiveAvgPool2d, nn.Upsample, nn.Hardtanh, HardQuant, nn.Linear, nn.LogSoftmax,
    nn.AvgPool2d, nn.Embedding, nn.Dropout, nn.ConstantPad1d, nn.Conv1d, nn.MaxPool1d, nn.GRU,
    nn.GRUCell, nn.Tanh, AvgPoolQuant, QuantizedHardtanh, AvgPool, Reorg, AdaptiveAvgPoolQuant, nn.LSTM,
    nn.LSTMCell, UpsampleQuant, WeightCenter, LinearQuant, nn.PixelShuffle, nn.LayerNorm,
    nn.UpsamplingBilinear2d, nn.LeakyReLU, ShuffleBlock, nn.InstanceNorm2d, ConvTranspose2dQuant,
    RoiAlign, RCConv2d, nn.ReflectionPad2d, ILN, H_Sigmoid, H_SigmoidQuant, H_Swish, H_SwishQuant, Cat,
    # drop op
    nn.Sigmoid, # SigmoidQuant.
    # new op
    AdaptiveAvgQuant, SigmoidOp, nn.ConstantPad2d, TanhOp, SoftMaxOp,
    TopK, Coords, PointSample, PointSampleQuant, BboxConvert, ROIAlign,
    nn.ELU, Swish, AdaptivePad, MaxOp, MeanOp, NormOp, Filter, nn.Unfold,
    DynamicConv2d, DynamicConvTranspose2d, ZerosOrOnes, Size, ScatterAdd,
    IndexSelect, CompareOp, BatchConv, SoftMaxAccurateOp, MinOp, GridSample2D,
    Clamp,
    # mask rcnn
    fpn_rpn_outputs, roi_2mlp_head, fast_rcnn_outputs, mask_rcnn_outputs, mask_rcnn_fcn_head_v1upXconvs
]

TWO_LAYERTYPES = [SliceMul, AvgQuant, MulQuant, Assign_Conv2D, SliceAddQuant]

THREE_LAYERTYPES = [AlphaFusion]

OTHER_LAYERS = [
    torch.cat, torch.mul, torch.bmm, torch.stack, torch.exp, torch.unique, torch.pow,
    torch.matmul,
    # torch.abs
]

REDUNDANCY_LAYER = [nn.GRU]

if USE_WIREFRAME:
    ONE_LAYERTYPES += [ProposalLines, LineJuncPick, LineJuncDecoder, LOIPooling]

#================================== data_flow
FIXED_ACTIVATION = [HardQuant, QuantizedHardtanh, SigmoidQuant, H_SigmoidQuant,
                    H_SwishQuant,
                   ]

FLOAT_ACTIVATION = [nn.Sigmoid, nn.Tanh]

MIXED_ACTIVATION = [SigmoidOp, TanhOp, SoftMaxOp]

NOT_COMBINE_ACTIVATION = [SigmoidOp]

#=================================================#
FLOAT_CONV = [nn.Conv2d, nn.ConvTranspose2d, nn.Linear, nn.Conv1d, RCConv2d, Assign_Conv2D
              ]

FIXED_CONV = [Conv2dQuant, LinearQuant, ConvTranspose2dQuant]

DATA_CONV = [nn.Conv2d, nn.ConvTranspose2d, nn.Linear, nn.Conv1d, RCConv2d,
             Conv2dQuant, LinearQuant, ConvTranspose2dQuant, Assign_Conv2D
             ]

# ycnn底层未将tanh_fixed和conv合并
IGNORE_ACTIVATION = [TanhOp]

#=================================================
DATA_OUT_FIXED = [
    HardQuant, AvgPoolQuant, QuantizedHardtanh, AdaptiveAvgPoolQuant,
    UpsampleQuant, H_SigmoidQuant, H_SwishQuant, AdaptiveAvgQuant, AvgQuant,
    SigmoidQuant, MulQuant, SliceAddQuant
]

DATA_OUT_FLOAT = [
    nn.AdaptiveAvgPool2d, nn.BatchNorm2d, nn.BatchNorm1d, nn.ReLU, nn.Upsample,
    nn.Hardtanh, nn.AvgPool2d, nn.GRU, nn.GRUCell, AvgPool, nn.LSTM,
    nn.LSTMCell, nn.UpsamplingBilinear2d, nn.LeakyReLU, nn.InstanceNorm2d,
    ILN, H_Sigmoid, H_Swish, nn.Sigmoid, nn.Tanh, nn.ELU, PixelMul, Swish,
    MulConstant, Sum, torch.bmm, BMM, ElementwiseAdd, SubConstant, ElementwiseSub,
    SliceDiv, Mul, BboxConvert, ROIAlign, MaxOp, MeanOp, NormOp, Exp, Filter,
    nn.LayerNorm, SliceAdd, nn.Unfold, CompareOp, DynamicConv2d,
    DynamicConvTranspose2d, ZerosOrOnes, ScatterAdd, Size,
    BatchConv, Pow, MatMul, Abs
]

DATA_OUT_MIXED = [
    nn.Softmax, nn.MaxPool2d, nn.LogSoftmax, nn.ConstantPad1d,
    nn.MaxPool1d, Reorg, WeightCenter, nn.PixelShuffle, ShuffleBlock,
    RoiAlign, nn.ReflectionPad2d, nn.ConstantPad2d, SliceMul,
    SlicePickRange,
    #
    torch.cat, Concat, Cat, View, torch.index_select, IndexSelect,
    #
    SigmoidOp, TanhOp, Permute, Scatter, Coords, PointSample, TopK,
    SoftMaxOp, Clone, AdaptivePad, AlphaFusion, SoftMaxAccurateOp,
    Repeat, Inverse, MinOp, GridSample2D, Split, ExpandAs
    ]

DATA_IN_FIXED = [AvgPoolQuant, UpsampleQuant, AvgQuant, SigmoidQuant, MulQuant, SliceAddQuant]

DATA_IN_FLOAT = [
    nn.AdaptiveAvgPool2d, nn.ReLU, nn.Upsample, nn.Hardtanh, nn.AvgPool2d,
    nn.Sigmoid, SigmoidOp, nn.Tanh, TanhOp, SoftMaxOp, HardQuant,
    nn.GRU, nn.GRUCell, AvgPool, nn.LSTM, nn.LSTMCell, nn.UpsamplingBilinear2d,
    nn.LeakyReLU, nn.InstanceNorm2d, ILN, H_Sigmoid, H_Swish, nn.BatchNorm1d, nn.BatchNorm2d,
    nn.ELU, PixelMul, Swish, MulConstant, Sum, torch.bmm, BMM, ElementwiseAdd, SubConstant,
    ElementwiseSub, SliceDiv, Mul, BboxConvert, ROIAlign, MaxOp, MeanOp, Exp, Filter,
    nn.LayerNorm, SliceAdd, nn.Unfold, CompareOp, DynamicConv2d, DynamicConvTranspose2d,
    ZerosOrOnes, ScatterAdd, Size, BatchConv, SoftMaxAccurateOp, Pow, MatMul, Abs
]

DATA_IN_MIXED = [
    nn.Softmax, nn.MaxPool2d, nn.LogSoftmax,
    nn.ConstantPad1d, nn.MaxPool1d, QuantizedHardtanh, Reorg, AdaptiveAvgPoolQuant,
    WeightCenter, nn.PixelShuffle, ShuffleBlock, RoiAlign, nn.ReflectionPad2d,
    H_SigmoidQuant, H_SwishQuant, AdaptiveAvgQuant, nn.ConstantPad2d, View,
    SliceMul, SlicePickRange,
    #
    torch.cat, Concat, Cat, Permute, Scatter, Coords, PointSample, TopK, Clone,
    AdaptivePad, AlphaFusion, NormOp, torch.unique, torch.index_select, IndexSelect,
    Repeat, Inverse, MinOp, GridSample2D, Split, ExpandAs
]

if USE_WIREFRAME:
    DATA_OUT_FLOAT += [ProposalLines, LineJuncPick, LineJuncDecoder, Unique, LOIPooling]
    DATA_IN_FLOAT += [ProposalLines, LineJuncPick, LineJuncDecoder, Unique, LOIPooling]

#=============================== for export_model
LAYER_ONE_MAX = [SoftMaxOp,
                 SigmoidQuant, SigmoidOp,
                 H_Sigmoid, H_SigmoidQuant]

LAYER_PASS_MAX = [Reshape, View]
### 以上为导模型内部使用标记，不作为训练使用 ###
version = "matting_v1_0"
key3 = [0x24, 0x8e, 0x27, 0x36, 0x61, 0x7e, 0xd3, 0x56, 0xa1, 0xf7, 0x15, 0x88, 0x19, 0xcf, 0x5f, 0x3c]
ekey3 = '248e2736617ed356a1f7158819cf5f3c'

# pytorch to ycnn: dim for tensor
dim_to_ycnn = {0: 2, 1: 0, 2: 1, None: -1}
# ================== for functions

flatten = lambda x: [y for l in x for y in flatten(l)] if type(x) is list or type(x) is tuple else [x]




